/* diagnostic and demo tool for keyboard encoding problems
 *
 * function keys are given to ncurses/terminfo for interpretation
 * (Linux only)
 *
 * normal keys are converted to Unicode according to the encoding
 * defined through locale charmap, Unicode Name is displayed 
 *
 * requires iculib International Components for Unicode 
 * (www.icu-project.org)  and ncurses (Linux only)
 * packages required for compile under Opensuse
 * ncurses-devel and libicu-devel 
 * compile with
 * g++ ushowkey.cpp -o ushowkey -lncurses -licuuc -licuio
 *
 * TODO error handling icu_error 
 *      read and convert stdin from file or pipe.
 *      implement/test surrogate pairs
 */

#if !defined(WIN32)
#include	<unistd.h>
#include	<termio.h>
#include	<langinfo.h>
#include        <ncurses.h>
#else
#include <windows.h>
#endif
#include	<stdlib.h>
#include	<stdio.h>
#include	<locale.h>
#include	<ctype.h>

#include <unicode/utypes.h>
#include <unicode/ustring.h>
#include <unicode/unistr.h>
#include <unicode/uchar.h>
#include <unicode/ucnv.h>

#include <unicode/ustdio.h>

#define DIM(x) (sizeof(x)/sizeof(x[0]))

#if !defined(WIN32)
struct termio	tty_mode;		/* original tty-mode is saved here */

int tty_raw(int fd);
int tty_reset(int fd);
void sprint_hex(unsigned char * phex, unsigned char * bytes, int bcount);

int buffered_read(int fd, void * buf, size_t buf_capacity);

#endif

int
main()
{
   int eof = FALSE;
   UChar uc[512];
   UFILE * u_stdout;
   char ucharName[1024];
   wchar_t kcode;
   char * szLocale = setlocale(LC_CTYPE, "");
   UErrorCode icu_error = U_ZERO_ERROR;

#if !defined(WIN32)
   unsigned char hex[4096];
   UChar * puc;
   unsigned char bytes[512];
   const char * pbytes;
   int32_t offsets[512];
   int  bcount, ucount, i;
   char * szTerm = getenv("TERM");
#else
   BOOL IsConsole;
   int retval, i;
   DWORD ucount;
   HANDLE hConsoleInput = GetStdHandle(STD_INPUT_HANDLE);
   HANDLE hConsoleOutput = GetStdHandle(STD_OUTPUT_HANDLE);
   DWORD dwConsMode;
   retval = GetConsoleMode(hConsoleInput, &dwConsMode);
   IsConsole = (retval == 1);
   if (!IsConsole)
   {
      printf("Can read from Console only\n");
      printf("Reading from File/Pipe not yet implemented\r\n");
      return(1);
   }
   dwConsMode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
   SetConsoleMode(hConsoleInput, dwConsMode);

#endif
   UConverter * ucnv_def;


#if !defined(WIN32)
   tty_raw(0);	/* save tty_mode */
   initscr();
   refresh();
   raw();
#endif

   u_stdout = u_finit(stdout, NULL, NULL);
   u_fprintf(u_stdout,"Press key(s) to display diagnostic info\r\n");
#if !defined(WIN32)
   u_fprintf(u_stdout,"check if key codes are in sync with locale encoding\r\n");
   u_fprintf(u_stdout,"and if function keys are recognized by ncurses/terminfo\r\n");
#else
   u_fprintf(u_stdout,"Best result with Lucida Console font\r\n");
#endif
   u_fprintf(u_stdout,"Leave with q, Ctrl-D, ESC, Return\r\n\r\n");

#if !defined(WIN32)
   u_fprintf(u_stdout,
            "Environment:\r\n\
	    \rTERM=%s    LC_CTYPE=%s    locale charmap %s\r\n",
            szTerm, (szLocale) ? szLocale : " !INVALID! ", nl_langinfo(CODESET));
#endif

   ucnv_def = ucnv_open(NULL, &icu_error);
   if (!ucnv_def)
   {
      u_fprintf(u_stdout,
                "\r\ncould not open default icu converter: %s\n\r",
		u_errorName(icu_error));
   }
   else
   {
//    icu_error = U_ZERO_ERROR;
//    u_fprintf(u_stdout, "using icu converter %s\r\n", ucnv_getName(ucnv_def, &icu_error));
   }




#if !defined(WIN32)
   u_fprintf(u_stdout, "\r\n%-5s%-10s%-30s%s\n\r",
             "Char", "Unicode", "raw hex sequence", "Name/Key" );
#else
   u_fprintf(u_stdout, "\r\n%-5s%-10s%s\n\r",
             "Char", "Unicode", "Name" );
#endif
   u_fflush(u_stdout);
   for (eof = FALSE; eof == FALSE; )
   {
#if !defined(WIN32)
       bcount = buffered_read(0, bytes, sizeof(bytes)-1);
       if (bcount == -1)
	    perror("read");
       if (bcount <= 0)
	    break;
       bytes[bcount] = 0;
#else
       retval = ReadConsoleW(hConsoleInput, uc, DIM(uc), &ucount, NULL);
#endif

#if !defined(WIN32)
      if (iscntrl(bytes[0]))
      {
         /* assume input buffer has a single f-key sequence      */
         /* expecting control code / function key sequence       */
	 /* from keyboard only (not from clipboard)              */

         /* display raw hex bytes */
	 sprint_hex(hex, bytes, bcount);
	 if (bcount == 1)
	 {
	    uc[0] = bytes[0];
	    u_fprintf(u_stdout, "%-5s", "");
	    u_fprintf(u_stdout, "U+%04x    ", uc[0]);

	 }
	 else
	 {
	    u_fprintf(u_stdout, "%-15s", "");
	 }
	 u_fprintf(u_stdout, "%-30s", hex);

	 /* control character: check for terminfo key definition */
	 keypad(stdscr, TRUE);
	 kcode = key_defined((const char*)bytes);
	 keypad(stdscr, FALSE);
	 if (kcode > 0)
	    u_fprintf(u_stdout, "%s", keyname(kcode));
	 else if (bcount == 1)
	 {   /* for a single control byte print Unicode Name */
	    icu_error = U_ZERO_ERROR;
	    u_charName(uc[0], U_EXTENDED_CHAR_NAME, ucharName, sizeof(ucharName), &icu_error);
	    u_fprintf(u_stdout, "%s", ucharName);
	 }
	 u_fprintf(u_stdout, "\n\r");
      }
      else
#endif
      {
	 /* sequence does not start with control code: convert to Unicode/UTF16 */
	 /* we may have received a char sequence from clipboard                 */

#if !defined(WIN32)
	 puc = uc;
	 pbytes = (const char *)bytes;
	 icu_error = U_ZERO_ERROR;
	 ucnv_reset(ucnv_def);
	 ucnv_toUnicode(ucnv_def, &puc, puc+DIM(uc)-1, &pbytes, pbytes+bcount, offsets, TRUE, &icu_error);
	 puc[0] = 0;

	 ucount = puc-uc;
	 offsets[ucount] = bcount;
#endif

#if 0
	icu_error = U_ZERO_ERROR;
	u_strToUTF32(uch, DIM(uch), NULL, uc, -1, &icu_error);
	if (u_isprint(uch[0]))
	{
	}
#endif

         for (i = 0; i < ucount; i++)
	 {
#if !defined(WIN32)
	    u_fputc(uc[i], u_stdout);
#else
            u_fflush(u_stdout);
	    if (u_isprint(uc[i]))
	    {
	       WriteConsoleW(hConsoleOutput, uc+i, 1, NULL, NULL);
	    }
	    else
	    {
	       WriteConsoleW(hConsoleOutput, L"\xfffd", 1, NULL, NULL);
	    }
#endif
	    u_fprintf(u_stdout, "%-4s", "");
	    u_fprintf(u_stdout, "U+%04x    ", uc[i]);
#if !defined(WIN32)
	    sprint_hex(hex, bytes+offsets[i], offsets[i+1]-offsets[i]);
	    u_fprintf(u_stdout, "%-30s", hex);
#endif
	    icu_error = U_ZERO_ERROR;
	    u_charName(uc[i], U_EXTENDED_CHAR_NAME, ucharName, sizeof(ucharName), &icu_error);
	    u_fprintf(u_stdout, "%s", ucharName);
	    u_fprintf(u_stdout, "\n\r");
	 }
      }
      u_fflush(u_stdout);
#if !defined(WIN32)
      kcode = bytes[0];
#else
      kcode = uc[0];
#endif
      switch(kcode)
      {
      case 3:
      case 4:
      case 13:
      case 26:
      case 'q':
      case 27:
#if !defined(WIN32)
        if (bcount == 1)
#endif
	   eof = TRUE;
	break;
      }
   }

#if !defined(WIN32)
   /* we do not call endwin because it would clear the screen */
   /* but we have to restore stty-modes                       */
   tty_reset(0);
#endif
    return(0);
}

#if !defined(WIN32)

#include	<sys/select.h>
int
buffered_read(int fd, void * buf, size_t buf_capacity)
{
   int retval, count;
   fd_set fdsReadList;
   struct timeval timTimeout;
   unsigned char * cbuf = (unsigned char*) buf;

   /* from some terminals (e.g. cygwin/XWin) we get one char per read */
   /* so wait 0.1 seconds to collect function key / utf8 sequence     */

   count = read(fd, cbuf, buf_capacity);
   if (count <= 0)
      return(count);

   do 
   {
      FD_ZERO(&fdsReadList);
      FD_SET(fd, &fdsReadList);
      timTimeout.tv_sec  = 0;	   /* seconds */
      timTimeout.tv_usec = 10000;  /* microseconds = 0.1 seconds */

      retval = select(NFDBITS, &fdsReadList, NULL, NULL, &timTimeout);
      if (retval == 0)
         return(count);
      if (retval == -1)
         return(-1);

      retval = read(fd, cbuf+count, buf_capacity-count);
      if (retval == 0)
         return(count);
      if (retval == -1)
         return(-1);
      count += retval;
   } while ((size_t)count < buf_capacity);
   return(count);

}
int
tty_raw(int fd)
{
	struct termio	temp_mode;

	if (ioctl(fd, TCGETA, (char*) &temp_mode) < 0)
		return (-1);

	tty_mode = temp_mode;		/* save tty-mode */

	temp_mode.c_iflag  = 0;
	temp_mode.c_oflag &= ~OPOST;
	temp_mode.c_lflag &= ~(ISIG | ICANON | ECHO | XCASE);
	temp_mode.c_cflag &= ~(CSIZE | PARENB);
	temp_mode.c_cflag |= CS8;

	temp_mode.c_cc[VMIN]	= 1;
	temp_mode.c_cc[VTIME]	= 20;

	if (ioctl(fd, TCSETAW, (char *)&temp_mode) < 0) 
		return (-1);
	return(0);
}
int
tty_reset(int fd)
{
	if (ioctl (fd, TCSETAW, (char *) &tty_mode) < 0)
		return (-1);
	return(0);
}
void
sprint_hex(unsigned char * phex, unsigned char * bytes, int bcount)
{
   int i;
   for (i = 0; i < bcount; i++)
   {
      if (i > 0)
      {
	 *phex++ = ' ';
      }
     phex += sprintf((char*) phex, "0x%02x", bytes[i]);
   }
}
#endif
